#include <windows.h>
#include <string.h>
#include <vorbis/vorbisenc.h>

#define PACKETBUFLEN 500

__declspec(dllexport) int vorbis_maxpacketlength = 500;

static vorbis_info      vi_enc, vi_dec;
static vorbis_dsp_state vd_enc, vd_dec;
static vorbis_block     vb_enc, vb_dec;
static ogg_packet       op_dec;
static unsigned char headerbyte_enc, headerbyte_dec, encinit;
static CRITICAL_SECTION cs;
static unsigned char packetbuffer_enc[PACKETBUFLEN];
static int packetbufferpos_enc, encodeloopcontinue;

static int vorbisbitrates[2][16] =
{8, 16, 20, 24, 28, 32, 40, 48, 56, 60, 64, 80, 96, 112, 128, 160,
12, 16, 24, 28, 32, 48, 56, 60, 64, 80, 96, 112, 128, 160, 192, 256};

static int vorbissfreq[8] = {8000, 12000, 16000, 22050, 24000, 32000, 44100, 48000};


// headerbyte: SFFFBBBB

__declspec(dllexport) int vorbis_init()
{
	InitializeCriticalSection(&cs);
	return 1;	// Version 1
}

__declspec(dllexport) int vorbis_closeenc()
{
	EnterCriticalSection(&cs);
	if(!encinit)
	{
		LeaveCriticalSection(&cs);
		return 0;
	}
	encinit = 0;
	vorbis_block_clear(&vb_enc);
	vorbis_dsp_clear(&vd_enc);
	vorbis_info_clear(&vi_enc);
	LeaveCriticalSection(&cs);
	return 0;
}

__declspec(dllexport) int vorbis_initenc(int channels, int sfreq, int bitrate)
{
	int i;

	EnterCriticalSection(&cs);
	if(encinit)
		vorbis_closeenc();
	vorbis_info_init(&vi_enc);
	if(sfreq == 44100)
		vorbis_encode_init(&vi_enc, channels, sfreq, bitrate, bitrate, -1);
	else vorbis_encode_init_vbr(&vi_enc, channels, sfreq, 0.0);
	vorbis_analysis_init(&vd_enc, &vi_enc);
	vorbis_block_init(&vd_enc, &vb_enc);
	headerbyte_enc = channels == 2 ? 128 : 0;
	for(i = 0; i < 8; i++)
		if(sfreq == vorbissfreq[i])
			headerbyte_enc |= i << 4;
	for(i = 0; i < 16; i++)
		if(bitrate / 1000 == vorbisbitrates[channels - 1][i])
			headerbyte_enc |= i;
	*packetbuffer_enc = headerbyte_enc;
	packetbufferpos_enc = 1;
	encodeloopcontinue = 0;
	encinit = 1;
	LeaveCriticalSection(&cs);
	return 0;
}

__declspec(dllexport) int vorbis_encframe(short *frame, int samples)
{
	float **buffer;
	int i;

	EnterCriticalSection(&cs);
	if(!encinit)
	{
		LeaveCriticalSection(&cs);
		return 0;
	}
	buffer = vorbis_analysis_buffer(&vd_enc, samples);
	if(vi_enc.channels == 2)
		for(i = 0; i < samples; i++)
		{
			buffer[0][i] = frame[i*2] / 32768.f;
			buffer[1][i] = frame[i*2+1] / 32768.f;
		}
	else for(i = 0; i < samples; i++)
		buffer[0][i] = frame[i] / 32768.f;
	vorbis_analysis_wrote(&vd_enc, samples);
	LeaveCriticalSection(&cs);
	return 0;
}

__declspec(dllexport) int vorbis_getpacket(unsigned char *packet)
{
	ogg_packet op;
	int retval;

	EnterCriticalSection(&cs);
	if(!encinit)
	{
		LeaveCriticalSection(&cs);
		return 0;
	}
	if(packetbufferpos_enc > 2 && !packetbuffer_enc[1])
	{
		memcpy(packet, packetbuffer_enc, packetbufferpos_enc);
		retval = packetbufferpos_enc;
		*packetbuffer_enc = headerbyte_enc;
		packetbufferpos_enc = 1;
		LeaveCriticalSection(&cs);
		return retval;
	}
	do {
		if(!encodeloopcontinue)
		{
			if(vorbis_analysis_blockout(&vd_enc, &vb_enc) != 1)
			{
				LeaveCriticalSection(&cs);
				return 0;
			}
			vorbis_analysis(&vb_enc, 0);
			vorbis_bitrate_addblock(&vb_enc);
		}
		retval = 0;
		encodeloopcontinue = 0;
		while(vorbis_bitrate_flushpacket(&vd_enc, &op))
		{
			if(op.bytes > 255)
			{
				memcpy(packet, packetbuffer_enc, packetbufferpos_enc);
				retval = packetbufferpos_enc;
				*packetbuffer_enc = headerbyte_enc;
				packetbuffer_enc[1] = 0;
				memcpy(packetbuffer_enc + 2, op.packet, op.bytes);
				packetbufferpos_enc = op.bytes + 2;
				encodeloopcontinue = 1;
				break;
			} else if(packetbufferpos_enc + 1 + op.bytes > vorbis_maxpacketlength)
			{
				memcpy(packet, packetbuffer_enc, packetbufferpos_enc);
				retval = packetbufferpos_enc;
				*packetbuffer_enc = headerbyte_enc;
				packetbuffer_enc[1] = (unsigned char)op.bytes;
				memcpy(packetbuffer_enc + 2, op.packet, op.bytes);
				packetbufferpos_enc = op.bytes + 2;
				encodeloopcontinue = 1;
				break;
			} else {
				packetbuffer_enc[packetbufferpos_enc] = (unsigned char)op.bytes;
				memcpy(packetbuffer_enc + packetbufferpos_enc + 1, op.packet, op.bytes);
				packetbufferpos_enc += 1 + op.bytes;
				retval = 0;
				encodeloopcontinue = 0;
			}
		}
	} while(!retval);
	LeaveCriticalSection(&cs);
	return retval;
}

__declspec(dllexport) int vorbis_decpacket(unsigned char *packet, int len)
{
	int channels, bitrate;

	EnterCriticalSection(&cs);
	if(headerbyte_dec != *packet)
	{
		ogg_packet header;
		ogg_packet header_comm;
		ogg_packet header_code;
		vorbis_comment vc_dec;

		if(headerbyte_dec)
		{
			vorbis_block_clear(&vb_dec);
			vorbis_dsp_clear(&vd_dec);
			vorbis_info_clear(&vi_dec);
		}
		headerbyte_dec = *packet;

		vorbis_info_init(&vi_dec);
		vorbis_comment_init(&vc_dec);
		channels = headerbyte_dec & 128 ? 2 : 1;
		bitrate = vorbisbitrates[channels - 1][headerbyte_dec & 0xf] * 1000;
		if(vorbissfreq[(headerbyte_dec & 0x70) >> 4] == 44100)
			vorbis_encode_init(&vi_dec, channels, vorbissfreq[(headerbyte_dec & 0x70) >> 4], bitrate, bitrate, -1);
		else vorbis_encode_init_vbr(&vi_dec, channels, vorbissfreq[(headerbyte_dec & 0x70) >> 4], 0.0);
		vorbis_analysis_init(&vd_dec, &vi_dec);
		vorbis_analysis_headerout(&vd_dec, &vc_dec, &header, &header_comm, &header_code);
		vorbis_synthesis_headerin(&vi_dec, &vc_dec, &header);
		vorbis_synthesis_headerin(&vi_dec, &vc_dec, &header_comm);
		vorbis_synthesis_headerin(&vi_dec, &vc_dec, &header_code);
		vorbis_dsp_clear(&vd_dec);
		vorbis_comment_clear(&vc_dec);

		vorbis_synthesis_init(&vd_dec, &vi_dec);
		vorbis_block_init(&vd_dec, &vb_dec);
		op_dec.packetno = 0;
		op_dec.b_o_s = 1;
		op_dec.e_o_s = 0;
	}
	if(!packet[1])
	{
		op_dec.packet = packet + 2;
		op_dec.bytes = len - 2;
		if(vorbis_synthesis(&vb_dec, &op_dec) == 0)
			vorbis_synthesis_blockin(&vd_dec, &vb_dec);
		op_dec.packetno++;
		op_dec.b_o_s = 0;
	} else {
		packet++;
		len--;
		while(len > 0)
		{
			op_dec.packet = packet + 1;
			op_dec.bytes = *packet;
			if(vorbis_synthesis(&vb_dec, &op_dec) == 0)
				vorbis_synthesis_blockin(&vd_dec, &vb_dec);
			op_dec.packetno++;
			op_dec.b_o_s = 0;
			len -= *packet + 1;
			packet += *packet + 1;
		}
	}
	LeaveCriticalSection(&cs);
	return 0;
}

__declspec(dllexport) int vorbis_getframe(short *frame, int samples)
{
	float **pcm;
	int samples_avail, i, val;

	if(!headerbyte_dec)
		return 0;
	EnterCriticalSection(&cs);
	samples_avail = vorbis_synthesis_pcmout(&vd_dec, &pcm);
	if(samples_avail < samples)
	{
		LeaveCriticalSection(&cs);
		return 0;
	}
	if(headerbyte_dec & 128)
		for(i = 0; i < samples; i++)
		{
			val = (int)(pcm[0][i]*32767.f);
			if(val > 32767) val = 32767;
			if(val < -32768) val = -32768;
			frame[i*2] = (short)val;
			val = (int)(pcm[1][i]*32767.f);
			if(val > 32767) val = 32767;
			if(val < -32768) val = -32768;
			frame[i*2+1] = (short)val;
		}
	else
		for(i = 0; i < samples; i++)
		{
			val = (int)(pcm[0][i]*32767.f);
			if(val > 32767) val = 32767;
			if(val < -32768) val = -32768;
			frame[i] = (short)val;
		}
	vorbis_synthesis_read(&vd_dec, samples);
	LeaveCriticalSection(&cs);
	return samples;
}

__declspec(dllexport) int vorbis_headerinfo(int header, int *channels, int *sfreq, int *bitrate)
{
	*channels = header & 128 ? 2 : 1;
	*sfreq = vorbissfreq[(header & 0x70) >> 4];
	*bitrate = vorbisbitrates[*channels - 1][header & 0xf];
	return 0;
}
